import pygame, random
from pygame.locals import *
import time

from math import copysign as sign

from screens.gameover import GameOver
from screens.text import TextOverlay

from enemies.enemy import JellyFish,Squid,Shark

size=32

levelmessages=["Got the idea?  Kill enemies to advance.",#1
"Building quickly can stop most things,\nbut not a shark",#2
"The squids like to try to find a way\naround your walls",#3
"Pink enemies take extra hits",#4
"Be careful with pink sharks,\n they can ruin your guns", #5
"Heating up a bit now",#6
"Orange means fast"
]

class Bubble:
	def __init__(self,game, target,x,y):
		self.target=target
		self.x=x
		self.y=y
		self.game=game
	def update(self):
		square=self.target.fromsquare
		if self.target.tosquare:
			square=self.target.tosquare
		if self.target not in self.game.enemies:
			self.game.remove_bubble(self)
			return
		realx=square.x*size
		realy=square.y*size+size
		deltax=realx-self.x
		deltay=realy-self.y
		
		if abs(deltax)+abs(deltay)<size:
			self.target.take_hit()
			self.game.remove_bubble(self)
		else:
			ymove=0
			xmove=0
			if deltax==0:
				ymove=sign(4,deltay)
			elif deltay==0:
				xmove=sign(4,deltax)
			elif deltax>deltay:
				xmove=sign(3,deltax)
				ymove=sign(2,deltay)
			else:
				ymove=sign(3,deltay)
				xmove=sign(2,deltax)
			self.x+=xmove
			self.y+=ymove
		
		

class Gun:
	easydirections={
	(1,0):0,
	(1,-1):1,
	(0,-1):2,
	(-1,-1):3,
	(-1,0):4,
	(-1,1):5,
	(0,1):6,
	(1,1):7,
	}
	def __init__(self,game):
		self.square=None
		self.target=None
		self.range=5
		self.game=game
		self.direction=0
		self.shootingprogress=0
	def explode(self):
		self.game.remove_gun(self)
	def update(self):
		if self.square and self.square.ischarged:
			if self.target and self.target in self.game.enemies:
				self.shootingprogress+=1
				self.direction=self.easydirections[self.dumbcmp(self.square.x,self.target.fromsquare.x),
														self.dumbcmp(self.square.y,self.target.fromsquare.y)]
				if self.shootingprogress>=30:
					self.game.bubbles.append(Bubble(self.game,self.target,self.square.x*size, self.square.y*size+size))
					random.choice(self.game.shootsounds).play()
					self.shootingprogress=1
				if self.range**2<(self.square.x-self.target.fromsquare.x)**2+(self.square.y-self.target.fromsquare.y)**2:
					self.target=None
			else:
				self.shootingprogress=0
				for enemy in self.game.enemies:
					if self.range**2>=(self.square.x-enemy.fromsquare.x)**2+(self.square.y-enemy.fromsquare.y)**2 and enemy.shootable:
						self.target=enemy
						self.shootingprogress=0
						break
	
	def dumbcmp(self, x1,x2):
		if(x1<x2):
			return 1
		if(x2<x1):
			return -1
		return 0
	
class GridSquare:
	def __init__(self,x,y):
		self.x=x
		self.y=y
		self.haswall=False
		self.ischargable=False
		self.ischarged=False
		self.gun=None
		self.ispowersource=False
		self.isblocked=False
		self.blockagetype=0
		self.haspearl=False
		self.enemy=None
	
	def update_from(self, other):
		self.haswall=other.haswall
		self.ischargable=other.ischargable
		self.gun=other.gun
		self.ispowersource=other.ispowersource
		self.isblocked=other.isblocked
		self.blockagetype=other.blockagetype
		self.haspearl=other.haspearl
		self.enemy=other.enemy
	def destroy(self):
		self.haswall=False
		self.ischargable=False
		if self.gun:
			self.gun.explode()
		self.gun=None
		self.ispowersource=False
		
	def has_something(self):
		return self.haswall or self.haspearl or self.enemy or self.isblocked
		
	def __lt__(self, other):
		return self.y< other.y or (self.y==other.y and self.x<other.x)
	def __gt__(self, other):
		return self.y > other.y or (self.y==other.y and self.x>other.x)
	def __eq__(self, other):
		return self.y== other.y and self.x==other.x
	def __le__(self, other):
		return sefl.__lt__(other) or self.__eq__(other)
	def __ge__(self, other):
		return self.__gt__( other) or self.__eq__(other)
	def __ne__(self, other):
		return not self.__eq__(other)
		
	def __hash__(self):
		return self.x.__hash__()+self.y.__hash__()

def image(path):
	return pygame.image.load('../images/%s'%path)

class GameScreen:
	
	def generate_piece(self):
		result=[[None for x in range(7)] for y in range(7)]
		squares=[GridSquare(0,0)]
		result[3][3]=squares[0]
		size=random.randint(3,5)
		while len(squares)<size:
			expandedsquare=random.choice(squares)
			pos=[expandedsquare.x+3,expandedsquare.y+3]
			if random.randint(0,1):
				valchange=1
			else:
				valchange=-1
			if random.randint(0,1):
				pos[0]+=valchange
			else:
				pos[1]+=valchange
			x,y=pos
			if not result[y][x]:
				added=GridSquare(x-3,y-3)
				squares.append(added)
				result[y][x]=added
		
		for square in squares:
			square.haswall=True
			if random.randint(1,5)>2:
				square.ischargable=True
				if random.randint(1,8)==8:
					square.gun=Gun(self)
				elif random.randint(1,7)==7:
					square.ispowersource=True
		return result
			

	def __init__(self, game, difficulty):
		self.game=game
		self.frame=0
		self.animationframe=0
		self.difficulty=difficulty
		self.mousegameX=0
		self.mousegameY=0
		
		self.points=0
		self.totalpoints=0#set this if you beat the level and create a new game scene
		self.pointstoadvance=difficulty**2+5
		
		self.frames_until_build=30
		
		self.frames_until_start=10*difficulty
		
		self.wall=image('block.png')
		self.ditheredwall=image('blockdithered.png')
		self.redwall=image('blockred.png')
		self.poweroff=image('poweroff.png')
		self.poweron=image('poweron.png'),image('poweron2.png')
		self.crystals=image('powercrystal.png'),image('powercrystal2.png')
		self.ditheredgun=image('turretdithered.png')
		self.redgun=image('turretred.png')
		self.ditheredcrystal=image('powercrystaldithered.png')
		self.redcrystal=image('powercrystalred.png')
		self.ditheredoff=image('poweroffdithered.png')
		self.redoff=image('poweroffred.png')
		self.pearl=image('pearl.png')
		self.bubble=image('bubble.png')
		
		self.barriers=image('seaweed1.png'),image('seaweed2.png'),image('coral1.png'),image('coral2.png')
		
		self.offturrets=[image('turretoff-000%i.png'%i) for i in range(8)]
		self.onturrets=[image('turreton-000%i.png'%i) for i in range(8)]
		self.shootingturrets=[ [image('turretshoot%i-000%i.png'%(i+1,j)) for i in range(2)] for j in range(8)]
		
		#self.turretshooting=[ [image('turretshoot%i-000%i.png'%(i,j)) for i in range(1,2)] for j in range(8)]
		
		self.shootsounds=[pygame.mixer.Sound('../sound/shoot%i.ogg'%i) for i in range(1,7)]
		self.buildsound=pygame.mixer.Sound('../sound/build.ogg')
		self.boomsound=pygame.mixer.Sound('../sound/destroy.ogg')
		self.pearltakensound=pygame.mixer.Sound('../sound/pearltaken.ogg')
		self.victory=pygame.mixer.Sound('../sound/victory.ogg')
		self.font=pygame.font.Font('../fonts/PressStart2P.ttf',16)
		
		self.background=pygame.Color( 0, 64, 88 )
		self.grid=[[GridSquare(x,y) for x in range(31)] for y in range(22)]
		
		self.isLegal=False
		
		self.guns=[]
		self.enemies=[]
		self.bubbles=[]
		
		
		
		#give them the initial 3 pearls:
		self.pearllocations=[(15,11),(16,11),(17,11)]
		self.grid[11][15].haspearl=True
		self.grid[11][16].haspearl=True
		self.grid[11][17].haspearl=True
		#surround those with walls
		self.grid[10][14].haswall=True
		self.grid[10][15].haswall=True
		self.grid[10][16].haswall=True
		self.grid[10][17].haswall=True
		self.grid[10][18].haswall=True
		self.grid[12][14].haswall=True
		self.grid[12][15].haswall=True
		self.grid[12][16].haswall=True
		self.grid[12][17].haswall=True
		self.grid[12][18].haswall=True
		self.grid[11][14].haswall=True
		self.grid[11][18].haswall=True
		

		self.edges=[]
		for x in range(31):
			self.edges.append( (x,0))
			self.edges.append( (x,21) )
			self.grid[0][x].isblocked=True
			self.grid[0][x].blockagetype=random.randint(0,3)
			self.grid[21][x].isblocked=True
			self.grid[21][x].blockagetype=random.randint(0,3)
		for y in range(22):
			self.edges.append( (0,y))
			self.edges.append( (30,y) )
			self.grid[y][0].isblocked=True
			self.grid[y][0].blockagetype=random.randint(0,3)
			self.grid[y][30].isblocked=True
			self.grid[y][30].blockagetype=random.randint(0,3)
		
		self.nextpiece=self.generate_piece()
		
		for i in range(self.difficulty):
			blockagetype=random.randint(0,3)
			ycoordinate=random.randint(1,len(self.grid)-1)
			xcoordinate=random.randint(1,len(self.grid[ycoordinate])-1)
			for j in range(self.difficulty//2):
				newx,newy=ycoordinate+random.randint(-j,j),xcoordinate+random.randint(-j,j)
				if newy>0 and newy<len(self.grid):
					if newx>0 and newx<len(self.grid[newy]):
						if not self.grid[newy][newx].has_something():
							self.grid[newy][newx].isblocked=True
							self.grid[newy][newx].blockagetype=blockagetype
		
	
	def take_pearl(self, square):
		square.haspearl=False
		self.pearltakensound.play()
		self.pearllocations.remove((square.x,square.y))
		self.enemies=[]
		if len(self.pearllocations)==0:
			self.game.active_screen=GameOver(self, self.game)
		for gridrow in self.grid:
			for square in gridrow:
				square.enemy=None
	
	def remove_bubble(self,bubble):
		if bubble in self.bubbles:
			self.bubbles.remove(bubble)
	
	def update(self):
		self.mousegameX=self.game.lastmousepos[0]//size
		self.mousegameY=self.game.lastmousepos[1]//size
		self.frame+=1
		self.animationframe=self.frame//5
		passes=True
		for gridrow in self.nextpiece:
			for gridsquare in gridrow:
				if gridsquare:
					x,y=self.mousegameX+gridsquare.x,self.mousegameY+gridsquare.y
					y=y-1#who cares why
					if y<0 or x<0 or y>=len(self.grid) or x >= len(self.grid[y]):
						passes=False
					elif self.grid[y][x].has_something():
						passes=False
		self.isLegal=passes
		
		if self.frame>self.frames_until_start:
			chance=self.difficulty+3
			roll=random.randint(0,300)
			while roll<chance:
				chance=chance-roll
				roll=random.randint(0,300)
				pos=random.choice(self.edges)
				color='blue'
				colorroll=random.randint(0,self.difficulty)
				if colorroll>7:
					color='orange'
				elif colorroll>3:
					color='pink'
				
				if random.randint(0,10)<8:
					self.enemies.append(JellyFish(self, pos[0],pos[1],color))
				elif random.randint(0,10)<5:
					self.enemies.append(Squid(self, pos[0],pos[1],color))
				else:
					self.enemies.append(Shark(self, pos[0],pos[1],color))
				
		
		for enemy in self.enemies:
			enemy.update()
		for gun in self.guns:
			gun.update()
		for bubble in self.bubbles:
			bubble.update()
		
		if not self.nextpiece:
			self.frames_until_build-=1
			if self.frames_until_build==0:
				self.frames_until_build=30
				self.nextpiece=self.generate_piece()
		
		if self.points>self.pointstoadvance:
			pygame.mixer.music.pause()
			
			next_level=GameScreen(self.game, self.difficulty+1)
			next_level.totalpoints=self.totalpoints
			message=["Level %i Complete!\nClick to continue."%self.difficulty]
			if self.difficulty-1<len(levelmessages):
				message+=[levelmessages[self.difficulty-1]]
			self.victory.play()
			time.sleep(3.5)
			pygame.mixer.music.unpause()
			self.game.active_screen=TextOverlay(self.game, next_level,message)
		
		
		
	def place_piece(self):
		self.buildsound.play()
		for gridrow in self.nextpiece:
			for gridsquare in gridrow:
				if gridsquare:
					x,y=self.mousegameX+gridsquare.x,self.mousegameY+gridsquare.y
					y=y-1#who cares why
					updated=self.grid[y][x]
					updated.update_from(gridsquare)
		#again!
		for gridrow in self.nextpiece:
			for gridsquare in gridrow:
				if gridsquare:
					x,y=self.mousegameX+gridsquare.x,self.mousegameY+gridsquare.y
					y=y-1#who cares why
					updated=self.grid[y][x]
					if updated.gun:
						updated.gun.square=updated
						self.guns.append(updated.gun)
					if updated.ispowersource:
						self.spread_positive(x,y)
					elif updated.ischargable:
						for neighbor in self.neighbors(x,y):
							square=self.grid[neighbor[1]][neighbor[0]]
							if square.ischarged:
								self.spread_positive(x,y)
								break
		self.nextpiece=[]

	def remove_gun(self, gun):
		self.guns.remove(gun)
	def inbounds(self, x,y=None):
		if y is None:
			x,y=x
		if x<0 or y<0:
			return False
		if y>=len(self.grid) or x>=len(self.grid[y]):
			return False
		return True
	
	def boom(self, x,y):
		self.boomsound.play()
		repair=set()
		for square in [ (x,y),(x+1,y),(x-1,y),(x,y+1),(x,y-1) ]:
			if self.inbounds(square):
				xv,yv=square
				repair=repair.union(self.spread_negative(xv,yv))
				
				self.grid[yv][xv].destroy()
		for item in repair:
			self.spread_positive(item[0],item[1])
	def smash(self, x,y):
		self.boomsound.play()
		repair=self.spread_negative(x,y)
		self.grid[y][x].destroy()
		for item in repair:
			self.spread_positive(item[0],item[1])
	
	def remove_enemy(self,arg):
		if arg in self.enemies:
			self.enemies.remove(arg)
	def add_enemy(self, enemy):
		self.enemies.append(enemy)
	
	def spread_positive(self, x,y):
		if x<0 or y<0 or y>=len(self.grid) or x>=len(self.grid[y]):
			return
		square=self.grid[y][x]
		if not square.ischargable:
			return
		if square.ischarged:
			return
		square.ischarged=True
		for neighbor in self.neighbors(x,y):
			self.spread_positive(neighbor[0],neighbor[1])
			
	def spread_negative(self, x,y):
		if x<0 or y<0 or y>=len(self.grid) or x>=len(self.grid[y]):
			return set()
		square=self.grid[y][x]
		if not square.ischargable:
			return set()
		if not square.ischarged:
			return set()
		square.ischarged=False
		if square.ispowersource:
			result=set ( ((x,y),) )
		else:
			result=set()
		for neighbor in self.neighbors(x,y):
			result=result.union(self.spread_negative(neighbor[0],neighbor[1]))
		return result
		#then when you've DONE that, you recharge from each power source
		
	
			
	def neighbors(self, x,y):
		return [m for m in [ (x-1,y),(x+1,y),(x,y-1),(x,y+1)] if self.inbounds(m)]
	def neighborsquares(self, square):
		return [self.grid[y][x] for x,y in self.neighbors(square.x,square.y)]
	
	def handle_event(self, event):
		if event.type==MOUSEBUTTONUP:
			if event.button==1:
				if self.isLegal:
					self.place_piece()
	
	
	
	def draw(self, window):
		window.fill(self.background)
		self.drawgrid(window)
		for enemy in self.enemies:
			enemy.draw(window)
		self.drawnextpiece(window)
		for bubble in self.bubbles:
			window.blit(self.bubble,  (bubble.x,bubble.y))
		self.drawUI(window)
	
	def drawUI(self, window):
		levelblob=self.font.render("Level %i"%self.difficulty,False,pygame.Color(252,252,252),self.background)
		levelrect=levelblob.get_rect()
		levelrect.topleft=(8,8)
		window.blit(levelblob,levelrect)
		
		scoreblob=self.font.render("Score: %i"%self.totalpoints,False,pygame.Color(252,252,252),self.background)
		scorerect=scoreblob.get_rect()
		scorerect.topleft=(208,8)
		window.blit(scoreblob,scorerect)
		
		nextblob=self.font.render("Next level: ",False,pygame.Color(252,252,252),self.background)
		nextrect=nextblob.get_rect()
		nextrect.topleft=(408,8)
		window.blit(nextblob,nextrect)
		
		nextmeteroutside=pygame.Rect(583,0,200,32)
		window.fill(pygame.Color(252,252,252),nextmeteroutside)

		
		
		emptywidth=196*(self.pointstoadvance-self.points)//self.pointstoadvance
		if emptywidth>196:
			emptywidth=196
		nextmeterinside=pygame.Rect(585+196-emptywidth,2,emptywidth,28)
		window.fill(self.background,nextmeterinside)
		
		
		
	def drawnextpiece(self,window):
		for gridrow in self.nextpiece:
			for gridsquare in gridrow:
				if gridsquare:
					pos=(self.mousegameX+gridsquare.x)*size ,(self.mousegameY+gridsquare.y)*size
					if gridsquare.haswall:
						if gridsquare.ischargable:
							if gridsquare.gun:
								if self.isLegal:
									window.blit(self.ditheredgun,pos)
								else:
									window.blit(self.redgun,pos)
								
							elif gridsquare.ispowersource:
								if self.isLegal:
									window.blit(self.ditheredcrystal,pos)
								else:
									window.blit(self.redcrystal,pos)
							else:
								if self.isLegal:
									window.blit(self.ditheredoff,pos)
								else:
									window.blit(self.redoff, pos)
						else:
							#draw dithered square
							if self.isLegal:
								window.blit(self.ditheredwall, pos)
							else:
								window.blit(self.redwall, pos)
		
	def drawgrid(self, window):
		for gridrow in self.grid:
			for gridsquare in gridrow:
				pos=gridsquare.x*size,gridsquare.y*size+size
				if gridsquare.haswall:
					if gridsquare.ischargable:
						if gridsquare.ischarged:
							if gridsquare.gun:
								if gridsquare.gun.shootingprogress>0:
									window.blit(self.shootingturrets[gridsquare.gun.direction][gridsquare.gun.shootingprogress//16],pos)
								else:
									window.blit(self.onturrets[gridsquare.gun.direction],pos)
							elif gridsquare.ispowersource:
								window.blit(self.crystals[self.animationframe%2],pos)
								pass
							else:
								window.blit(self.poweron[self.animationframe%2],pos)
						else:
							if gridsquare.gun:
								window.blit(self.offturrets[gridsquare.gun.direction],pos)
							else:
								window.blit(self.poweroff,pos)
								pass
					else:
						window.blit(self.wall, pos)
				elif gridsquare.isblocked:
					window.blit(self.barriers[gridsquare.blockagetype],pos)
				elif gridsquare.haspearl:
					window.blit(self.pearl,pos)
					
	def get_music(self):
		musics=["background"]
		return musics[self.difficulty%len(musics)]