import pygame
from pygame.locals import *
from screens.game import GameScreen
from screens.tutorial import Tutorial

class MenuItem:
	def __init__(self, text):
		self.text=text
		self.rect=None
		self.selected=False
		self.hovered=None

class Menu:
	def __init__(self, game):
		self.menuoptions=["Resume","New Game","Tutorial","Exit"]
		self.menuoptions=[MenuItem(x) for x in self.menuoptions]
		self.game=game
		self.menufont=pygame.font.Font('../fonts/PressStart2P.ttf',16*3)
		self.enabledcolor=pygame.Color(100,160,120)
		self.selectedcolor=pygame.Color(100,200,120)
		self.disabledcolor=pygame.Color(50,80,60)
		self.backgroundcolor=pygame.Color( 0, 64, 88 )
		
		self.selection_sound=pygame.mixer.Sound('../sound/menu_select.ogg')
		self.confirm_sound=pygame.mixer.Sound('../sound/menu_confirm.ogg')
		
		menuX=150
		menuY=30
		for item in self.menuoptions:
			item.rect=self.menufont.render(item.text,False, self.enabledcolor).get_rect()
			item.rect.topleft=(menuX,menuY)
			menuY=item.rect.bottom+30
		
	def update(self):
		pass
	def handle_event(self, event):
		if event.type==MOUSEMOTION:
			pos=event.pos
			self.hovered=None
			for item in self.menuoptions:
				wasselected=item.selected
				item.selected=item.rect.collidepoint(pos)
				if item.selected:
					self.hovered=item
				if item.selected and not wasselected and self.isEnabled(item):
					self.selection_sound.play()
		elif event.type==MOUSEBUTTONUP:
			if event.button==1:
				self.do_selected()
	def do_selected(self):
		if self.hovered and self.isEnabled(self.hovered):
			self.confirm_sound.play()
			if self.hovered.text=='Resume':
				self.game.active_screen=self.game.paused_screen
				self.game.paused_screen=None
			elif self.hovered.text=="New Game":
				self.game.active_screen=GameScreen(self.game, 1)
				self.game.paused_screen=None
			elif self.hovered.text=="Tutorial":
				self.game.active_screen=Tutorial(self.game)
				self.game.paused_screen=None
			elif self.hovered.text=="Exit":
				pygame.event.post(pygame.event.Event(QUIT))
	def isEnabled(self,menu):
		if menu.text=="Resume":
			return self.game.paused_screen
		return True
	def draw(self, window):
		window.fill(self.backgroundcolor)
		for item in self.menuoptions:
			color=self.disabledcolor
			if self.isEnabled(item):
				color=self.enabledcolor
				if item.selected:
					color=self.selectedcolor
			blob=self.menufont.render(item.text,False, color,self.backgroundcolor)
			window.blit(blob,item.rect)
	def get_music(self):
		return None

